/* -*- Mode: C++ -*- */

/* COPYING
 * Liquid Rendering Library
 * (c) 2009 by Henrik Theiling
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Known Non-Conformance, Glitches, Peculiarities, or other Differences:
 *    - The presentation forms cannot currently be used in the input
 *      strings.  Their use is strongly deprecated by Unicode, so this
 *      does not seem to bad.  Use base forms + ZWJ/ZWNJ instead.
 *
 *      The reason is that our fonts strictly distinguish glyphs from
 *      code points, so they do not need code points to store glyph
 *      variations as the four joining forms of Arabic.  Therefore,
 *      it would be a waste of memory to duplicate all Arabic glyphs
 *      in their presentation form code points.
 *
 *      Further, although we do need code points for ligatures, it would
 *      be wrong to 'reuse' presentation form code points for those that
 *      are missing a neutral form.  E.g., there is no LAM+ALEF ligature
 *      in Unicode, but only presentation forms thereof.  We hide the
 *      LAM+ALEF code point in a part of the font that is outside of
 *      Unicode instead.
 *
 *      If we ever support presentation forms, we will have a separate
 *      module that decomposes them (protecting them from their environment
 *      with ZWNJers.  E.g.
 *             <LAM+ALEF FINAL FORM>
 *          -> <ZWNJ> <LAM> <ALEF> <ZWJ> <ZWNJ>
 *
 * TODO:
 *    - Reset use_space before spaces, ZWNJ, etc.
 *
 *    - The normal space is 1px too wide for Arabic.  We need more
 *      context so that we can define another space for Arabic (and
 *      probably other cursive scripts).  E.g. let each glyph define
 *      a script ID that it matched (in logical order) in glyph
 *      selection.
 *
 *    - The full stop is on a different base line in Arabic, because
 *      the Latin and the Arabic base lines are different.  We should
 *      have contextual forms for the full stop in Arabic mode (and for
 *      other punctuaton with the same problem, of course).
 *
 *    - Remove (a single) BOM from beginning of string if present.
 *
 *    - Composition and decomposition: must be done before ligature
 *      selection and should know all canonical decompositions.  Maybe
 *      this should be made optional, because some might already have
 *      the data in NFC form.
 *      Note: we won't do NFC, but NFC-like, font-driven composition.
 *      And probably no decomposition at all.
 *
 *    - Code point restructuring: see render.c for TODO.
 *      - ligatures [DONE]
 *      - composition (base + NSM): needed for Lithuanian, Ming-Dong, and
 *        many other languages where NFC still contains NSMs.
 *
 *    - Anchors, e.g. 26 ones called a..z or A..Z.  These could be placed
 *      directly in the bitmap, a..z instead of . and A..Z instead of #.
 *      (Problem: two anchors at the same position.)
 *      If these work, we could probably remove a lot of pre-composed
 *      base + NSM (removal could be done automatically, of course).  This
 *      would probably need NFD/NFC, however, because the user couldn't
 *      know what form to use to get the output right.
 *
 *    - language support ('lang' attribute).  Fonts already have it,
 *      but there's no API yet to set the language of a given text.
 *
 *      Maybe we need control codes for changing the language, although it
 *      is most likely that the whole text will use the same language (for
 *      the font), even if it contains several languages.  So we'll start
 *      with a global language setting.
 *
 *    - Maybe implement a more efficient storage of pixel bitmaps in
 *      ftx.pl, by reducing the font to only 4 columns (a great majority
 *      of chars has max. 4 non-zero columns).  One way to start would be
 *      to only store the columns x1...x2.   The current 31kb could
 *      be reduced to 26kB.
 *
 *      We can also squeeze the code point ID by splitting it into 8 bit
 *      offset + 8 (and probably less) bit page index into a map of page
 *      index to page base.  Liqmean currently has chars in only 28 pages,
 *      so instead of 21 bits, we'd only need 13 (= 8 + 5).  This means we
 *      could store 5 columns of matrix bits and still only need 64 bits
 *      per (small) glyph (at least with current settings).
 *
 *    - Allow specifying glyphs with a width of 0 pixels.
 *
 * DONE:
 *    - Default ignorable: do it right (from derivedcoreproperties).
 *      Currently, even ZWJ generate a glyph.
 *      Also note that we remove all BN characters during the bidi step,
 *      which is not right.  Some do have a shape, but only need to be
 *      ignored for bidi.
 *      DONE: - CharInfo::ignorable is correctly set up in liquid.c.
 *              It only needs to be used consistently now!
 *            - Glyph and ligature selection ignore them
 *            - Auto-Joining and explicit joining ignore them
 *      FIXME: check that we missed nothing
 */

#ifndef LIQUID_INTERNAL_H_10923821098321
#define LIQUID_INTERNAL_H_10923821098321

#define IN_LIQUID_COMPILING 1

#include <stdio.h>
#include <string.h>
#include <liquid/render++.h>
#include <liquid/font.h>

#ifndef LIQ_CALLOC
#define LIQ_CALLOC   calloc
#endif
#ifndef LIQ_FREE
#define LIQ_FREE     free
#endif

#ifndef INLINE
#  if defined(LIQ_INLINE_0)
#    define INLINE inline
#  elif defined(LIQ_INLINE_1)
#    define INLINE _inline
#  elif defined(LIQ_INLINE_2)
#    define INLINE __inline
#  elif defined(LIQ_INLINE_2_2)
#    define INLINE __inline__
#  elif defined(_MSC_VER) || defined(__WATCOMC__)
#    define INLINE _inline
#  elif defined(__GNUC__)
#    define INLINE __inline__
#  else
#    define INLINE
#  endif
#endif

/* ******************************************************************************** */

#define UNICODE_CHAR(X)                ((LiqChar32)X##UL)

#define UNICODE_REPLACEMENT_CHARACTER  UNICODE_CHAR(0x00fffd)
#define UNICODE_VARIATION_SELECTOR_1   UNICODE_CHAR(0x00fe00)
#define UNICODE_VARIATION_SELECTOR_16  UNICODE_CHAR(0x00fe0f)
#define UNICODE_VARIATION_SELECTOR_17  UNICODE_CHAR(0x0e0100)
#define UNICODE_VARIATION_SELECTOR_256 UNICODE_CHAR(0x0e01ef)
#define UNICODE_ZWJ                    UNICODE_CHAR(0x00200d)
#define UNICODE_ZWNJ                   UNICODE_CHAR(0x00200c)
#define UNICODE_MISSING_GLYPH          UNICODE_CHAR(0x0f0600) /* private use */

/* ******************************************************************************** */

#define UTF_ACCEPT 0
#define UTF_REJECT 1

/* ******************************************************************************** */

#define unless(x)    if(!(x))
#define forever      for(;;)

/* ******************************************************************************** */

/* Bits 0 and 1 reserved for neutral bidi class */
typedef enum _LiqJoining {
    LIQ_NO_JOIN          = 0,

    /* Preseve information from ZWJ and ZWNJ */
    /* Note: after bidi, left and right might need swapping */
    /* If bit numbers are changed here, liq_bidi_swap_explicit_joiners has to be changed, too. */
    /* Also note that JOIN_LEFT and JOIN_RIGHT must have the same bit values as the
     * corresponding LIQ_GLYPH_* constants. */
    LIQ_JOIN_LEFT        = 0x01, /* 0=automatic, 1=do join left */
    LIQ_JOIN_RIGHT       = 0x02, /* 0=automatic, 1=do join right */
    LIQ_NO_JOIN_LEFT     = 0x04, /* 0=automatic, 1=don't join left */
    LIQ_NO_JOIN_RIGHT    = 0x08, /* 0=automatic, 1=don't join right */
                                 /* if both JOIN and NO_JOIN flags are set, it means
                                  * normal joining is applied, but ligatures are
                                  * prohibited.  This can be requested by using the
                                  * sequence ZWJ ZWNJ ZWJ between the characters. */

    LIQ_JOIN_MASK_LEFT   = LIQ_JOIN_LEFT  | LIQ_NO_JOIN_LEFT,
    LIQ_JOIN_MASK_RIGHT  = LIQ_JOIN_RIGHT | LIQ_NO_JOIN_RIGHT,
    LIQ_JOIN_MASK_GLYPH  = LIQ_JOIN_LEFT  | LIQ_JOIN_RIGHT
} LiqJoining;

static INLINE LiqJoining liq_joining_swap_left_right (LiqJoining x)
{
    return ((x & LIQ_JOIN_MASK_LEFT) << 1) | ((x & LIQ_JOIN_MASK_RIGHT) >> 1);
}

typedef enum _LiqBreakType {
    LIQ_BREAK_XX      = 0,
    LIQ_BREAK_BK      = 1
} LiqBreakType;

typedef enum _LiqBreakAction {
    LIQ_BREAK_NONE        = 0,
    LIQ_BREAK_BEFORE      = 1,
    LIQ_BREAK_AFTER       = 2,
    LIQ_BREAK_AND_DROP    = LIQ_BREAK_BEFORE | LIQ_BREAK_AFTER
} LiqBreakAction;

typedef enum _LiqBidiType {
    /* If more than 10 bits are needed, you must change BidiRange.
     *
     * With 10 bits, values may be between 0x000 and 0x3ff, i.e., the bit
     * value is 0x200.
     */

    /* convenience marker bits for fast checking: */
    LIQ_BIDI_NEUTRAL   = 0x00, /* must be 0 so that the neutrals are >= 3 */
    LIQ_BIDI_WEAK      = 0x04, /* must be 4 so that LIQ_BIDI_NSM is  >= 5 */
    LIQ_BIDI_STRONG    = 0x08,

    LIQ_BIDI_NUMBER    = 0x10,
    LIQ_BIDI_STRONG_L  = 0x20,
    LIQ_BIDI_STRONG_R  = 0x40,
    LIQ_BIDI_STRONG_LR = LIQ_BIDI_STRONG_L | LIQ_BIDI_STRONG_R,
    LIQ_BIDI_SEP_TERM  = 0x80,

    LIQ_BIDI_B_OR_S    = 0x02,

    /* Actual BIDI class values: the first column is for making the value
     * unique. When more marker bits are added, less such values might
     * be needed .*/
    LIQ_BIDI_L         = 0x00 | LIQ_BIDI_STRONG | LIQ_BIDI_STRONG_L,
    LIQ_BIDI_R         = 0x00 | LIQ_BIDI_STRONG | LIQ_BIDI_STRONG_R,
    LIQ_BIDI_AL        = 0x01 | LIQ_BIDI_STRONG | LIQ_BIDI_STRONG_R,
    LIQ_BIDI_LRE       = 0x00 | LIQ_BIDI_STRONG,
    LIQ_BIDI_LRO       = 0x01 | LIQ_BIDI_STRONG,
    LIQ_BIDI_RLE       = 0x02 | LIQ_BIDI_STRONG,
    LIQ_BIDI_RLO       = 0x03 | LIQ_BIDI_STRONG,

    LIQ_BIDI_EN        = 0x00 | LIQ_BIDI_WEAK   | LIQ_BIDI_NUMBER,
    LIQ_BIDI_AN        = 0x01 | LIQ_BIDI_WEAK   | LIQ_BIDI_NUMBER,
    LIQ_BIDI_ES        = 0x00 | LIQ_BIDI_WEAK   | LIQ_BIDI_SEP_TERM,
    LIQ_BIDI_ET        = 0x01 | LIQ_BIDI_WEAK   | LIQ_BIDI_SEP_TERM,
    LIQ_BIDI_CS        = 0x02 | LIQ_BIDI_WEAK   | LIQ_BIDI_SEP_TERM,
    LIQ_BIDI_PDF       = 0x02 | LIQ_BIDI_WEAK,    /* must not be part of flags, because B_OR_S */
    LIQ_BIDI_BN        = 0x03 | LIQ_BIDI_WEAK,    /* is set. */

    /* the following are <= 5 and thus copied in the flags for the Lx rules: */
    LIQ_BIDI_NSM       = 0x00 | LIQ_BIDI_WEAK,    /* must not have flag B_OR_S! */

    LIQ_BIDI_ON        = 0x00,                    /* must be 0 because B,S,WS must not be */
    LIQ_BIDI_WS        = 0x01,                    /* the flags for use in L1 rule. */
    LIQ_BIDI_B         = 0x00 | LIQ_BIDI_B_OR_S,  /* These must have values 0..3 so */
    LIQ_BIDI_S         = 0x01 | LIQ_BIDI_B_OR_S   /* they can redundantly be stored in */
} LiqBidiType;

#define LIQ_ORIG_BIDI_MAX 0x05                    /* 6 and 7 are disallowed because they have the
                                                   * B_OR_S bit set, which is checked separately */

#define LIQ_KERN_MIN  -16
#define LIQ_KERN_MAX  +15

/* Info about characters/glyphs: will be modified by the rendering steps */
struct _LiqCharInfo {
    /* Size is more important than speed in this library, and bit masking is
     * very fast anyway.   However, this structure is > 32 bits in any case,
     * so it's to 64 bits by the compiler.  Therefore, we make some fields
     * larger than necessary for faster access.  If we're about to hit the
     * 64 bit limit, we'll rethink.
     * We also try to avoid unnecessary byte/word boundaries for fields
     * smaller than the alignment of the boundaries, e.g. orig_bidi, kerning,
     * flags do not cross byte boundaries.
     */
    LIQ_UBITS32 code          : 32; /* 21 bits minimum (should be enough for glyphs, too) */

    LIQ_UBITS8  bidi          : 8;  /* 8 bits minimum */
    LIQ_UBITS8  level         : 6;  /* 6 bits minimum */
    LIQ_UBITS8  line_break    : 2;  /* line break action (not line break type!) */

    LIQ_UBITS8  orig_bidi     : 3;  /* only stores S,B,WS,NSM, all others are mapped to ON */
    LIQ_SBITS8  kern          : 5;  /* kern= -16..+15
                                     *
                                     * Kerning is stored for the right side of a glyph, i.e.,
                                     * the left glyph or an adjacent pair stores the kerning
                                     * for the pair, and the last glyph in a line is guaranteed
                                     * to have a kerning of 0, because no glyph follows.
                                     */

    LIQ_UBITS8  joining       : 4;  /* 4 bits minimum */

    LIQ_UBITS8  opaque        : 1;  /* Unset for all transparent glyphs by the bidi
                                     * algorith, for the arabicshaping algorithm.  Set for all
                                     * chars that need to be considered. */
    LIQ_UBITS8  mirror        : 1;  /* Set by glyph selector for mirroring iff no
                                     * mirrored glyph was found, so code is still the
                                     * original glyph
                                     * Also used temporarily in algorithm in Nx rules */
    LIQ_UBITS8  is_glyph      : 1;  /* false: code is a code point; true: code is a glyph id.
                                     * This is for consistency checks only: the code should
                                     * know at all times whether code is a code point or
                                     * a glyph id */
    LIQ_UBITS8  important     : 1;  /* bool: negative of Default_Ignorable_Code_Point */
};

struct _LiqParagraphInfo {
    LiqCharInfo      *first;
    LiqCharInfo      *last;
    int               level;
};

struct _LiqLineInfo {
    LiqParagraphInfo *paragraph;
    LiqCharInfo      *first;
    LiqCharInfo      *last;
    size_t            pixel_start;
    size_t            pixel_width;
};


#define LIQ_CHAR_INFO_0  {                    \
            0,               /* code */       \
            LIQ_BIDI_BN,     /* bidi */       \
            0,               /* level */      \
            LIQ_BREAK_NONE,  /* line_break */ \
            LIQ_BIDI_ON,     /* orig_bidi */  \
            0,               /* kern */       \
            LIQ_NO_JOIN,     /* joining */    \
            LIQ_FALSE,       /* opaque */     \
            LIQ_FALSE,       /* mirror */     \
            LIQ_FALSE,       /* is_glyph */   \
            LIQ_FALSE        /* important */  \
        }

extern LiqCharInfo const liq_char_info_0;

extern LiqCalloc liq_calloc;
extern LiqFree   liq_free;

/* ******************************************************************************** */

static INLINE int max_int (int a, int b)
{
    return a > b ? a : b;
}

static INLINE int min_int (int a, int b)
{
    return a < b ? a : b;
}

static INLINE LiqBool is_even (int i)
{
    return !(i & 1);
}

static INLINE LiqBool is_odd (int i)
{
    return !is_even(i);
}

static INLINE LiqBidiType level2bidi (int level)
{
    return is_odd(level) ? LIQ_BIDI_R : LIQ_BIDI_L;
}

#define TCALLOC(T,N) ((T*)liq_calloc(N,sizeof(T)))
#define TFREE(X)     ((liq_free((X))), (X)=NULL)

/* ********************************************************************** */

extern void        liq_split_paragraphs      (LiqRender *self);
extern void        liq_split_lines           (LiqRender *self);
                                             
extern void        liq_set_info              (LiqCharInfo *, LiqChar32);
extern void        liq_render_alloc_info     (LiqRender *, size_t);
                                             
extern void        liq_seq_reverse           (LiqCharInfo *, LiqCharInfo *);
extern void        liq_para_select_glyphs    (LiqRender *, LiqParagraphInfo *);
extern void        liq_para_select_ligatures (LiqRender *, LiqParagraphInfo *);

extern void        liq_run_auto_join         (LiqRender *,
                                              LiqParagraphInfo *,
                                              int,
                                              LiqBidiType,
                                              LiqBidiType,
                                              LiqCharInfo *,
                                              LiqCharInfo *);

extern void        liq_render_finalise_text  (LiqRender *);

extern void liq_foreach_run (
    LiqRender    *self,
    LiqParagraphInfo *para,
    void          (*step) (
                      LiqRender    *self,
                      LiqParagraphInfo *para,
                      int           run_level,
                      LiqBidiType   sor,
                      LiqBidiType   eor,
                      LiqCharInfo  *first,
                      LiqCharInfo  *last));

/* ********************************************************************** */

#endif /* LIQUID_INTERNAL_H_10923821098321 */
